/**
 * @file
 *
 * Partitioning stage of partitioned hash join
 *
 * @author Jens Teubner <jens.teubner@cs.tu-dortmund.de>
 *
 * $Id$
 */

#include "config.h"

#include "join/partition.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <pthread.h>

#include "hashtable/hashtable.h"
#include "task_queue.h"

/**
 * number of tuples to put into one buffer entry (when
 * using software-managed buffers
 */
#define BUFFER_SIZE      8

/**
 * Mask to implement modulo in a cheaper way.
 */
#define BUFFER_SIZE_MASK 7

#define WITH_SOFTWARE_MANAGED_BUFFERS 1

/**
 * A context of this type is passed to a thread that will compute
 * a local histogram (pthreads only allows to pass one pointer to
 * the thread in pthread_create()).
 */
struct comp_hist_ctx_t {
    tuple_t        *tuples;
    unsigned long   count;
    unsigned int    from_bit;
    unsigned int    to_bit;
    unsigned long  *ret;
};

/**
 * A context of this type is passed to a thread that will partition
 * a part of the input relation.
 */
struct part_ctx_t {
    tuple_t        *src;       /**< beginning of assigned input slice */
    unsigned long   count;     /**< number of tuples that should be processed */
    unsigned int    from_bit;
    unsigned int    to_bit;
    tuple_t       **dst;       /**< array with pointers to beginning of each
                                    output partition */
};

/*
 * The first round of partitioning is globally coordinated (do parallel
 * histograms and prefix sum, then run a number of partitioning threads).
 *
 * From then on, threads will refine partitions, and they will take a
 * partition, compute a local histogram, then refine the partition.
 */
struct refine_part_ctx_t {
    tuple_t        *src;      /**< beginning of the assigned partition */
    unsigned long   count;    /**< size of assigned partition */
    unsigned int    from_bit;
    unsigned int    to_bit;
    tuple_t       **dst;      /**< dst[0] tells where the partitioned output
                                   shall be put; all the other dst[i] are
                                   determined based on the histogram.
                                   This array must have been allocated from
                                   outside and dst[0] set; remaining elements
                                   are written from here */
};
typedef struct refine_part_ctx_t refine_part_ctx_t;


/** Wrapper for compute_histogram(), to cooperate with pthreads. */
static void * comp_hist_thrd (void *ctx);

static void
compute_histogram (const tuple_t *tuples, unsigned long count,
                   const unsigned int from_bit, const unsigned int to_bit,
                   unsigned long ret[]);

/**
 * Worker for partition().
 */
static void
partition_worker (const unsigned long num_tuples,
                  const unsigned int from_bit, const unsigned int to_bit,
                  const tuple_t *src, tuple_t **dst)
{
    const unsigned long  num_partitions = 1 << (to_bit - from_bit +1);
    const unsigned long  mask     = num_partitions - 1;
    const unsigned int   shift_by = sizeof (unsigned long) * 8 - to_bit;
    unsigned long       *pos;

#if WITH_SOFTWARE_MANAGED_BUFFERS
    tuple_t             *buf;
#endif

    pos = malloc (num_partitions * sizeof (*pos));
    assert (pos);

#if WITH_SOFTWARE_MANAGED_BUFFERS
    buf = malloc (BUFFER_SIZE * num_partitions * sizeof (*buf));
    assert (buf);
#endif

    /* initialize histogram */
    for (unsigned long i = 0; i < num_partitions; i++)
        pos[i] = 0;

    /* finally: partition */
    for (unsigned long i = 0; i < num_tuples; i++)
    {
        unsigned long hash = (HASH(src[i].key) >> shift_by) & mask;

#if WITH_SOFTWARE_MANAGED_BUFFERS

        /*
        printf ("copying %3u|%10u to buf[%3lu] (hash: %lu).\n",
                src[i].key, src[i].value,
                hash * BUFFER_SIZE + (pos[hash] & BUFFER_SIZE_MASK), hash);
        */

        buf[hash * BUFFER_SIZE + (pos[hash] & BUFFER_SIZE_MASK)] = src[i];
        pos[hash]++;

        /* flush out data at cache line boundaries */
        if (((unsigned long) (dst[hash] + pos[hash]) & 63) == 0)
        {
            /*
            printf ("flushing buffer with hash value %lu.\n", hash);
            */

            for (unsigned int j
                    = (pos[hash] > BUFFER_SIZE) ? pos[hash] - BUFFER_SIZE : 0;
                    j < pos[hash]; j++)
            {
                /*
                printf ("copying from buf[%3lu] to dst[%3lu][%3u].\n",
                        hash * BUFFER_SIZE + (j & BUFFER_SIZE_MASK),
                        hash, j);
                */

                dst[hash][j] = buf[hash * BUFFER_SIZE + (j & BUFFER_SIZE_MASK)];
            }
        }

#else
        dst[hash][pos[hash]] = src[i];
        pos[hash]++;
#endif
    }

#if WITH_SOFTWARE_MANAGED_BUFFERS

    /* flush out remaining data from buf[] */
    for (unsigned long i = 0; i < num_partitions; i++)
    {
        if (((unsigned long) (dst[i] + pos[i]) & 63) != 0)
        {
            /*
            printf ("flushing buffer with hash value %lu.\n", i);
            */

            for (unsigned int j
                    = (pos[i] > BUFFER_SIZE) ? pos[i] - BUFFER_SIZE : 0;
                    j < pos[i]; j++)
            {
                /*
                printf ("copying from buf[%3lu] to dst[%3lu][%3u].\n",
                        i * BUFFER_SIZE + (j & BUFFER_SIZE_MASK),
                        i, j);
                */

                dst[i][j] = buf[i * BUFFER_SIZE + (j & BUFFER_SIZE_MASK)];
            }
        }
    }

#endif

    free (pos);
}


static void *
part_worker_thrd (void *ctx)
{
    struct part_ctx_t *part_ctx = (struct part_ctx_t *) ctx;

    partition_worker (part_ctx->count, part_ctx->from_bit, part_ctx->to_bit,
                      part_ctx->src, part_ctx->dst);

    return NULL;
}

static void *
refine_partition_thrd (void *queue)
{
    task_queue_t      *q = (task_queue_t *) queue;
    refine_part_ctx_t *task;
    unsigned long     *histogram = NULL;

    /*
    fprintf (stderr, "starting thread %p\n", pthread_self ());
    */

    while ((task = task_queue_trydeq (q)))
    {
        /*
        fprintf (stderr, "thread %p: task is src=%p, count=%lu\n",
                pthread_self (), task->src, task->count);
        */

        const unsigned long fanout = 1 << (task->to_bit - task->from_bit + 1);

        histogram = realloc (histogram, fanout * sizeof (*histogram));
        assert (histogram);

        compute_histogram (
                task->src, task->count, task->from_bit, task->to_bit,
                histogram);

        for (unsigned int i = 1; i < fanout; i++)
            task->dst[i] = task->dst[i-1] + histogram[i-1];

        /*
        fprintf (stderr, "thread %p: dst[0]=%p, dst[1]=%p, dst[2]=%p\n",
                pthread_self (), task->dst[0], task->dst[1], task->dst[2]);
        */

        partition_worker (
                task->count, task->from_bit, task->to_bit,
                task->src, task->dst);
    }

    return NULL;
}

/**
 * Takes a partitioned relation and partitions it further.
 */
partitioned_relation_t
partition (const partitioned_relation_t rel,
           const unsigned int from_bit, const unsigned int to_bit)
{
    partitioned_relation_t  ret;
    unsigned long           fanout = (1 << (to_bit - from_bit + 1));

    ret.num_tuples = rel.num_tuples;
    ret.num_part   = rel.num_part * fanout;
    ret.startaddrs = malloc ((ret.num_part + 1) * sizeof (*ret.startaddrs));

    if (!ret.startaddrs)
    {
        fprintf (stderr,
                "Could not allocate memory for .startaddrs in partition().\n");
        exit (EXIT_FAILURE);
    }

    ret.startaddrs[0] = malloc (rel.num_tuples * sizeof (**ret.startaddrs));

    if (!ret.startaddrs)
    {
        fprintf (stderr, "Could not allocate memory for .startaddrs[0] "
                "in partition().\n");
        exit (EXIT_FAILURE);
    }

    for (unsigned long i = 1; i <= rel.num_part; i++)
        ret.startaddrs[i * fanout]
            = ret.startaddrs[0] + (rel.startaddrs[i] - rel.startaddrs[0]);

    for (unsigned int i = 0; i < rel.num_part; i++)
    {
        /*
         * Compute target addresses where partitions are going to be put.
         *
         * To this end,
         *
         *  1. Compute a histogram (to know sizes of each partition)
         *
         *  2. Populate ret.startaddrs accordingly
         */
        unsigned long       *histogram;
        const unsigned long  num_partitions = 1 << (to_bit - from_bit +1);

        histogram = malloc (num_partitions * sizeof (*histogram));
        assert (histogram);

        /* compute histogram */
        compute_histogram (
                rel.startaddrs[i], rel.startaddrs[i+1] - rel.startaddrs[i],
                from_bit, to_bit, histogram);

        /* determine beginning of each partition with help of prefix sum */
        assert (ret.startaddrs[i * fanout]);
        for (unsigned long j = 1; j < num_partitions; j++)
            ret.startaddrs[i * fanout + j]
                = ret.startaddrs[i * fanout + j-1] + histogram[j-1];

        free (histogram);

        /* now do the actual partitioning */
        partition_worker (
                rel.startaddrs[i+1] - rel.startaddrs[i], from_bit, to_bit,
                rel.startaddrs[i], ret.startaddrs + i * fanout);
    }

    return ret;
}

/*
 * Worker to compute a histogram for a (partial) relation
 */
static void
compute_histogram (const tuple_t *tuples, unsigned long count,
                   const unsigned int from_bit, const unsigned int to_bit,
                   unsigned long ret[])
{
    const unsigned long  fanout = (1 << (to_bit - from_bit + 1));
    const unsigned long  mask     = fanout - 1;
    const unsigned int   shift_by = sizeof (unsigned long) * 8 - to_bit;

    for (unsigned int i = 0; i < fanout; i++)
        ret[i] = 0;

    for (unsigned int i = 0; i < count; i++)
    {
        const unsigned long hash
            = (HASH(tuples[i].key) >> shift_by) & mask;
        ret[hash]++;
    }
}

static void *
comp_hist_thrd (void *ctx)
{
    struct comp_hist_ctx_t *comp_hist_ctx = (struct comp_hist_ctx_t *) ctx;

    /*
    fprintf (stderr,
            "Running comp_hist_thrd with tuples = %p, count = %lu, "
            "from_bit = %u, ret = %p\n", comp_hist_ctx->tuples,
            comp_hist_ctx->count, comp_hist_ctx->from_bit, comp_hist_ctx->ret);
    */

    compute_histogram (comp_hist_ctx->tuples, comp_hist_ctx->count,
                       comp_hist_ctx->from_bit, comp_hist_ctx->to_bit,
                       comp_hist_ctx->ret);

    return NULL;
}

/**
 * Takes a partitioned relation and partitions it further.
 */
partitioned_relation_t
parallel_partition (const partitioned_relation_t rel,
                    const unsigned int from_bit, const unsigned int to_bit,
                    const unsigned int parallelism_degree)
{
    const unsigned long  fanout = (1 << (to_bit - from_bit + 1));

    /* array to hold thread ids */
    pthread_t *threads;
    threads = malloc (parallelism_degree * sizeof (*threads));
    assert (threads);

    if (rel.num_part == 1)
    {
        /*
         * This is the first stage of partitioning.
         *
         * Thus:
         *
         *  1. Compute a histogram in parallel (by computing sub-histograms
         *     and summing them up)
         *
         *  2. Partition in parallel (after computing start addresses for
         *     each thread and partition.
         */

        /* prepare histogram data structures (to be populated by threads) */
        unsigned long **histograms;

        histograms = malloc (parallelism_degree * sizeof (*histograms));
        assert (histograms);

        for (unsigned int i = 0; i < parallelism_degree; i++)
        {
            histograms[i] = malloc (fanout * sizeof (*histograms[i]));
            assert (histograms[i]);
        }

        /* create a thread context for each thread and spawn a thread */
        struct comp_hist_ctx_t *comp_hist_ctxs;

        comp_hist_ctxs = malloc (parallelism_degree * sizeof (*comp_hist_ctxs));
        assert (comp_hist_ctxs);

        /* spawn all the threads */
        for (unsigned int i = 0; i < parallelism_degree; i++)
        {
            comp_hist_ctxs[i] = (struct comp_hist_ctx_t) {
                .tuples = rel.startaddrs[0]
                    + i * (rel.num_tuples / parallelism_degree),
                .count  = (i == parallelism_degree - 1)
                    ? (rel.num_tuples
                            - (parallelism_degree - 1) * (rel.num_tuples / parallelism_degree))
                    : (rel.num_tuples / parallelism_degree),
                .from_bit = from_bit,
                .to_bit = to_bit,
                .ret = histograms[i]
            };

            if (pthread_create (
                    &threads[i], NULL, comp_hist_thrd, &comp_hist_ctxs[i]))
            {
                fprintf (stderr, "Could not spawn thread %u.\n", i);
                exit (EXIT_FAILURE);
            }
        }

        /* wait for threads to complete */
        for (unsigned int i = 0; i < parallelism_degree; i++)
            if (pthread_join (threads[i], NULL))
            {
                fprintf (stderr, "Could not join thread %u.\n", i);
                exit (EXIT_FAILURE);
            }

        unsigned long *sum_histogram;

        sum_histogram = malloc (fanout * sizeof (*sum_histogram));
        assert (sum_histogram);

        /* initialize sum_histogram */
        for (unsigned int i = 0; i < fanout; i++)
            sum_histogram[i] = 0;

        /* add up values computed by threads */
        for (unsigned int i = 0; i < parallelism_degree; i++)
            for (unsigned int j = 0; j < fanout; j++)
                sum_histogram[j] += histograms[i][j];

        /*
         * We now have a complete histogram.  A prefix sum of this
         * complete histogram tells us where all the partitions
         * (on the global level) should go.
         */

        /* allocate memory for the partitioned relation
         * (which we'll later return) */
        partitioned_relation_t  ret;
        
        ret.num_tuples = rel.num_tuples;
        ret.num_part   = fanout;

        /* one more start-of-partition pointer than we have partitions;
         * last one points just after the last partition */
        ret.startaddrs = malloc ((fanout + 1) * sizeof (*ret.startaddrs));

        if (!ret.startaddrs)
        {
            fprintf (stderr,
                    "Could not allocate memory for .startaddrs in partition().\n");
            exit (EXIT_FAILURE);
        }

        /* allocate memory for the actual relation */
        ret.startaddrs[0] = malloc (rel.num_tuples * sizeof (**ret.startaddrs));

        if (!ret.startaddrs)
        {
            fprintf (stderr, "Could not allocate memory for .startaddrs[0] "
                    "in partition().\n");
            exit (EXIT_FAILURE);
        }

        /* this is that last pointer, just after the last partition */
        ret.startaddrs[fanout] = ret.startaddrs[0] + rel.num_tuples;

        /* determine beginning of each partition with help of prefix sum */
        for (unsigned long j = 1; j < fanout; j++)
            ret.startaddrs[j] = ret.startaddrs[j-1] + sum_histogram[j-1];

        free (sum_histogram);

        /* set up thread contexts for partitioning threads */
        struct part_ctx_t *part_ctx;
        part_ctx = malloc (parallelism_degree * sizeof (*part_ctx));
        assert (part_ctx);

        /* set up first context */
        part_ctx[0] = (struct part_ctx_t) {
            .src      = comp_hist_ctxs[0].tuples,
            .count    = comp_hist_ctxs[0].count,
            .from_bit = from_bit,
            .to_bit   = to_bit,
            .dst      = malloc (fanout * sizeof (*part_ctx[0].dst))
        };

        for (unsigned int i = 0; i < fanout; i++)
            part_ctx[0].dst[i] = ret.startaddrs[i];

        /* then all the rest, each one based on the previous */
        for (unsigned int i = 1; i < parallelism_degree; i++)
        {
            part_ctx[i] = (struct part_ctx_t) {
                .src      = comp_hist_ctxs[i].tuples,
                .count    = comp_hist_ctxs[i].count,
                .from_bit = from_bit,
                .to_bit   = to_bit,
                .dst      = malloc (fanout * sizeof (*part_ctx[i].dst))
            };

            for (unsigned int j = 0; j < fanout; j++)
                part_ctx[i].dst[j]
                    = part_ctx[i-1].dst[j] + comp_hist_ctxs[i-1].ret[j];
        }

        for (unsigned int i = 0; i < parallelism_degree; i++)
        {
            if (pthread_create (
                        &threads[i], NULL, part_worker_thrd, &part_ctx[i]))
            {
                fprintf (stderr, "Error spawning partitioner thread\n");
                exit (EXIT_FAILURE);
            }
        }

        /* wait for threads to complete */
        for (unsigned int i = 0; i < parallelism_degree; i++)
            if (pthread_join (threads[i], NULL))
            {
                fprintf (stderr, "Could not join thread %u.\n", i);
                exit (EXIT_FAILURE);
            }

        return ret;
    }
    else
    {
        /*
         * Create a number of partitioning tasks, put them into
         * a task queue, and spawn a set of threads.
         */
        partitioned_relation_t  ret;

        /*
         * Allocate memory for partitioned result; pre-set where all
         * the refined versions of each existing partition starts.
         */
        ret.num_tuples = rel.num_tuples;
        ret.num_part   = rel.num_part * fanout;
        ret.startaddrs = malloc ((ret.num_part + 1) * sizeof (*ret.startaddrs));

        if (!ret.startaddrs)
        {
            fprintf (stderr,
                    "Could not allocate memory for .startaddrs.\n");
            exit (EXIT_FAILURE);
        }

        ret.startaddrs[0] = malloc (rel.num_tuples * sizeof (**ret.startaddrs));

        if (!ret.startaddrs)
        {
            fprintf (stderr, "Could not allocate memory for .startaddrs[0].\n");
            exit (EXIT_FAILURE);
        }

        for (unsigned long i = 1; i <= rel.num_part; i++)
            ret.startaddrs[i * fanout]
                = ret.startaddrs[0] + (rel.startaddrs[i] - rel.startaddrs[0]);

        /*
        fprintf (stderr, "starting another round of partitioning...\n");
        */

        /*
         * Allocate a number of tasks to work on, plus a task queue.
         */
        refine_part_ctx_t  *ref_part_ctx;
        task_queue_t       *queue;

        ref_part_ctx = malloc (rel.num_part * sizeof (*ref_part_ctx));
        assert (ref_part_ctx);

        queue = task_queue_init (rel.num_part + 2);

        for (unsigned long i = 0; i < rel.num_part; i++)
        {
            /* define the task */
            ref_part_ctx[i] = (struct refine_part_ctx_t) {
                .src      = rel.startaddrs[i],
                .count    = rel.startaddrs[i+1] - rel.startaddrs[i],
                .from_bit = from_bit,
                .to_bit   = to_bit,
                .dst      = &ret.startaddrs[i * fanout]
            };

            /*
            fprintf (stderr, "enqueing task %lu...\n", i);
            */

            /* and enqueue it */
            task_queue_enq (queue, &ref_part_ctx[i]);
        }

        /*
        fprintf (stderr, "%lu tasks queued into %p\n", rel.num_part, queue);
        */

        for (unsigned int i = 0; i < parallelism_degree; i++)
            pthread_create (&threads[i], NULL, refine_partition_thrd, queue);

        /*
        fprintf (stderr, "%u threads spawned\n", parallelism_degree);
        */

        for (unsigned int i = 0; i < parallelism_degree; i++)
            pthread_join (threads[i], NULL);

        /*
        fprintf (stderr, "%u threads joined\n", parallelism_degree);
        */

        return ret;

        /* return partition (rel, from_bit, to_bit); */
    }
}
